// -*- java -*-

package eem;

import robocode.*;
import robocode.util.Utils;
import java.util.ArrayList; 
import java.util.*;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;

import eem.external.trees.secondGenKD.KdTree;

/**
 * zapper - robocode robot by Eugeniy E. Mikhailov
 */
public class zapper extends AdvancedRobot {
	static boolean firstTimeInit = true;
	static boolean isBulletShielder = false;
	double gunSkippingFireWhenReadyCnt = 0;
	double bulletEnergy=3;
	double firingAngle=0;
	double gunPrecision=3./180.*Math.PI;
	long time = 0;
	long enemyLastSeenTime = 0;
	long enemyUnseenGracePeriod = 10;
	botStatus enemy = null;
	botStatus bot   = null;//new botStatus();
	Point BattleField = new Point();
	double BattleFieldDiagonal = 0;
	Point SWcorner = null;
	Point NEcorner = null;
	Point SWcornerCollision = null;
	Point NEcornerCollision = null;
	Bullet _bullet=null;
	Driver _driver = null;
	double lastEnemyBulletEnergy = 2; // 2 is good default for a typical enemy bullet
	boolean virtrualHitOnRealWave = true; // shall we track for seen by enemy angles in virtual waves
	boolean enemyEmitsVirtualWaves = true; // shall enemy emit a wave every click
	static long firedBulletCnt=0;
	static long hitBulletCnt=0;
	static long enemyFiredBulletCnt=0;
	static long enemyHitBulletCnt=0;
	static long myWaveCnt=0;    // these two are to be able decay the old data
	static long enemyWaveCnt=0;
	double myStatsDecayTime = 1000;
	public ArrayList<Wave> myWaves = new ArrayList<Wave>();
	public ArrayList<Wave> enemyWaves = new ArrayList<Wave>();
	double ramAvoidDistance = 200;

	int neigborsNumForMasterBotGun = 100;
	int neigborsNumForEnemyBotGun = 10;
	int myGunKdTreeSize = 100000;
	int myGunKdTreeDims = 5;
	int MaxVirtualWavesMy = myGunKdTreeSize;
	int MaxVirtualWavesEnemy = myGunKdTreeSize; // Math.min(500, myGunKdTreeSize/3);
	public static KdTree.WeightedManhattan<Hit> hitsByMeTree = null;
	public static KdTree.WeightedManhattan<Hit> realWaveHitsByMeTree = null;
	public static KdTree.WeightedManhattan<Hit> hitsByEnemyTree = null;
	public static KdTree.WeightedManhattan<Hit> realWaveHitsByEnemyTree = null;

	// stats
	public static HashMap<String, GunStats> gunStatsMap = null;
	public static int[] notMatchedFiringSolutionCnt = null;
	public static int[] notFiredWhenCold = null;
	public static int[] bulletHitBulletCnt = null;
	public static int winCnt = 0;

	// gun managing related
	double IDEAL_BULLET_ENERGY = 1.95;
	HashMap<String, Double> fSols = new HashMap<String, Double>();
	String bestGunName = "NaN";


	public void run() {
		// start of the Round
		setColors(Color.red, Color.white, Color.red);
		BattleField.x = getBattleFieldWidth();
		BattleField.y = getBattleFieldHeight();
		BattleFieldDiagonal = BattleField.dist( new Point() );
		SWcorner = new Point(18,18);
		SWcornerCollision = new Point(19,19);
		NEcorner = new Point(BattleField.x-18, BattleField.y-18);
		NEcornerCollision = new Point(BattleField.x-19, BattleField.y-19);
		Driver _driverRC = new RandomCircle();
		Driver _driverWS = new WaveSurfer();
		_driver = _driverWS;
		myWaves = new ArrayList<Wave>();
		enemyWaves = new ArrayList<Wave>();
		enemyLastSeenTime = 0;;
		if ( hitsByMeTree == null ) {
			hitsByMeTree = new KdTree.WeightedManhattan<Hit>(myGunKdTreeDims, myGunKdTreeSize);
		}
		if ( realWaveHitsByMeTree == null ) {
			realWaveHitsByMeTree = new KdTree.WeightedManhattan<Hit>(myGunKdTreeDims, myGunKdTreeSize);
		}
		if ( hitsByEnemyTree == null ) {
			hitsByEnemyTree = new KdTree.WeightedManhattan<Hit>(myGunKdTreeDims, myGunKdTreeSize);
		}
		if ( realWaveHitsByEnemyTree == null ) {
			realWaveHitsByEnemyTree = new KdTree.WeightedManhattan<Hit>(myGunKdTreeDims, myGunKdTreeSize);
		}
		if ( gunStatsMap == null ) {
			gunStatsMap = new HashMap<String, GunStats>();
			GunStats _gs;
		}
		if (firstTimeInit) {
			firstTimeInit = false;
			notMatchedFiringSolutionCnt = new int[getNumRounds()];
			notFiredWhenCold = new int[getNumRounds()];
			bulletHitBulletCnt = new int[getNumRounds()];
		}

		setAdjustGunForRobotTurn(true);
		setAdjustRadarForGunTurn(true);

		setTurnRadarRight(Double.POSITIVE_INFINITY);

		while (true) {
			initTic();
			if ( (time - enemyLastSeenTime) > enemyUnseenGracePeriod ) {
				if (getOthers() > 0 ) { dbg("RADAR SLIP. SHAME! IT SHOULD NOT HAPPEN"); }
				setTurnRadarRight(Double.POSITIVE_INFINITY);
			}

			_driver = _driverWS;
			if ( enemy != null ) {
				if (enemy.pos.dist( bot.pos ) < ramAvoidDistance ) {
					// we are possibly rammed
					// switch to random motion with good anti ram
					//_driver = _driverRC;
				}
			}
			_driver.move();

			execute();
		}

	}

	
	public class Driver {
		Point destinationPoint = new Point(BattleField.x/2, BattleField.y/2);
		public void move(){}
	}

	public class RandomCircle extends Driver {
		double circlingDirection = 1;
		double timeToChangeDirection = 0;
		public void move(){
			Point c = new Point();
			double R = 200; //BattleFieldDiagonal / 2;
			if ( enemy == null ) {
				// move to center
				this.destinationPoint.x = BattleField.x/2;
				this.destinationPoint.y = BattleField.y/2;
				moveToDestination( this.destinationPoint );
			} else {
				// simple circle around enemy in random directions
				double headOnAngle = enemy.pos.angleTo( bot.pos ); // angle from enemy
				double a = headOnAngle;
				double desiredDist = BattleFieldDiagonal/3;
				double distToEnemy = bot.pos.dist(enemy.pos);
				boolean enemyIsRamming = (distToEnemy <= ramAvoidDistance);
				timeToChangeDirection--;
				//if ( timeToChangeDirection < 0 && !enemyIsRamming ) {
				if ( timeToChangeDirection < 0 ) {
					//dbg("changing direction");
					this.circlingDirection *= -1;;
					timeToChangeDirection = Math.random()*40;
					if (  Math.random() < .1 ) {
						//dbg("long leap");
						timeToChangeDirection =  Math.random()*20;
					}
				}
				double aOffset = Math.PI/2*(distToEnemy-30)/desiredDist;
				if ( distToEnemy < 55) { aOffset = 0; }
				a += this.circlingDirection*aOffset;
				//dbg("aOf " + aOffset );
				double da = Math.PI/10;
				da = this.circlingDirection*da;
				double dR = 10;
				//R=desiredDist + dR;
				Point destP = new Point();
				//dbg("looking for new destination");
				//a += this.circlingDirection*aOffset;
				int cnt=0, cntMax=100;
				dbg("dist to enemy: "+ distToEnemy);
				boolean isDirectionFlipped = false;
				R = 4+Math.abs(bot.speed)/8*R;
				do {
					//R -= dR;
					destP.x = bot.pos.x + R*Math.cos(a);
					destP.y = bot.pos.y + R*Math.sin(a);
					//dbg( "R = " + R + " " + "da = " +da + " " +  destP );
					a += this.circlingDirection*da;
					if ( enemyIsRamming && Math.abs(a-headOnAngle)> Math.PI/2 ) {
						if ( !isDirectionFlipped ) {
							dbg("Dive in protection, attempted to move toward ramming enemy");
							this.circlingDirection *= -1;
							a=headOnAngle;
							isDirectionFlipped = true;
						} else {
							dbg("we are cornered, attempting to ram back");
							a=headOnAngle + Math.PI;
							destP.x = bot.pos.x + R*Math.cos(a);
							destP.y = bot.pos.y + R*Math.sin(a);
							break;
						}
					}
				} while ( !destP.isInsideBox( SWcorner, NEcorner) && (cntMax > cnt++)  );
				dbg( "dest = " +  destP );
				this.destinationPoint.x = destP.x;
				this.destinationPoint.y = destP.y;

				moveToDestination( this.destinationPoint );
			}
		}
	}

	public class WaveSurfer extends Driver {
		Point destinationPoint = new Point();
		public void move() {
			Wave wToSurf = null;
			double waveClosestDist = Double.NEGATIVE_INFINITY;
			boolean virtualWave = false;
			for (int i = 0; i < enemyWaves.size(); i++) {
				Wave w = (Wave) enemyWaves.get(i);
				if (!w.isReal) continue; // not surfing virtual waves
				double surfDist = w.getSurfingDistanceTo(bot.pos);
				// positives mean that wave is behind the point
				if (surfDist > -2 ) {
					// this wave either behind or we are surfing it
					// so too late to worry about it
					continue;
				} else {
					if ( surfDist > waveClosestDist ) {
						// this wave is closer to us
						waveClosestDist = surfDist;
						wToSurf = w;
					}
				}

			}
			if (wToSurf == null && enemy != null) {
				// we create a fake wave
				wToSurf = new Wave();
				wToSurf.source = enemy;
				wToSurf.target = bot;
				// TODO a typical enemy bullet energy
				wToSurf.bulletEnergy = 3.0;
				assignWaveDanger( wToSurf );
				virtualWave = true;
			}
			if (wToSurf == null) {
				//dbg("No enemy waves to surf");
				// do some default motion
				return;
			}
			if (wToSurf.surfingInfo.isCalculated) {
				// if we have it calculated, we already moving there
				// also it helps against default bot alignment (0 rad) 
				// when remaining travel distance is 0
				//return;
			}
			double maxExcapeAnge = maxEscapeAngleForSpeed(wToSurf.getSpeed());
			double safestDanger = Double.POSITIVE_INFINITY;
			double headOnAngle = wToSurf.source.pos.angleTo( wToSurf.target.pos );
			double refAngle = wToSurf.source.pos.angleTo( bot.pos );
			// find how far bot could move CW or CCW from current position
			// before it get hit
			// positives mean that wave is behind the point
			int timeToSurf = (int) Math.round(-wToSurf.getSurfingDistanceTo(bot.pos)/wToSurf.getSpeed());
			double distToEnemy = wToSurf.source.pos.dist( bot.pos );
			int maxSurfPrdictionTime = 20;
			int minSurfPrdictionTime = 5;
			if (virtualWave) minSurfPrdictionTime =10;
			if (distToEnemy < 100 ) minSurfPrdictionTime=5;
			timeToSurf = Math.min( timeToSurf, maxSurfPrdictionTime );
			timeToSurf = Math.max( timeToSurf, minSurfPrdictionTime );

			int safestBin =0;

			//dbg("timeToSurf = " + timeToSurf);

			// checking CCW direction
			int surfDirection = 1; // 1 means CCW, -1 means CW
			DangerPoint ccwDestinationPoint = findTheSafestPontOnWave( wToSurf, timeToSurf, surfDirection );
			firingAngle = Utils.normalRelativeAngle( wToSurf.source.pos.angleTo( ccwDestinationPoint.pos ) - headOnAngle );
			int gfBin = xToBinInRange( firingAngle, -maxExcapeAnge, maxExcapeAnge, wToSurf.Nbins );
			//dbg("safest CCW point " +  ccwDestinationPoint + " and gfBin " + gfBin);
			if ( ccwDestinationPoint.danger < safestDanger ) {
				safestDanger = ccwDestinationPoint.danger;
				safestBin = gfBin;
				destinationPoint = ccwDestinationPoint.pos;
			}

			// checking CW direction
			surfDirection = -1; // 1 means CCW, -1 means CW
			DangerPoint cwDestinationPoint = findTheSafestPontOnWave( wToSurf, timeToSurf, surfDirection );
			firingAngle = Utils.normalRelativeAngle( wToSurf.source.pos.angleTo( cwDestinationPoint.pos ) - headOnAngle );
			gfBin = xToBinInRange( firingAngle, -maxExcapeAnge, maxExcapeAnge, wToSurf.Nbins );
			//dbg("safest CW  point " +  cwDestinationPoint + " and gfBin " + gfBin);
			if ( cwDestinationPoint.danger < safestDanger ) {
				safestDanger = cwDestinationPoint.danger;
				safestBin = gfBin;
				destinationPoint = cwDestinationPoint.pos;
			}

			//dbg("" + arrayToTextPlot(wToSurf.waveDanger) + " <- wave danger");
			firingAngle = Utils.normalRelativeAngle( wToSurf.source.pos.angleTo( destinationPoint ) - headOnAngle );
			double[] myPosArray = new double[ wToSurf.Nbins ];
			double myAngle = wToSurf.source.pos.angleTo( bot.pos );
			int myCurrentBin=xToBinInRange( myAngle - headOnAngle, -maxExcapeAnge, maxExcapeAnge, wToSurf.Nbins );
			myPosArray[ myCurrentBin ]=-3;
			myPosArray[ safestBin ]=5;
			//dbg("" + arrayToTextPlot(myPosArray) + " <- my current and reachable positions in the wave");
			//dbg("========================================================");
			wToSurf.surfingInfo.safestReachableBin = safestBin;
			wToSurf.surfingInfo.destinationPoint = this.destinationPoint;
			wToSurf.surfingInfo.isCalculated = true;
			moveToDestination( this.destinationPoint );
		}
	}

	public DangerPoint findTheSafestPontOnWave(Wave wToSurf, int timeToSurf, double surfDirection ) {
		Point destP = new Point( bot.pos );
		double safestDanger = Double.POSITIVE_INFINITY;
		double desiredDist = BattleFieldDiagonal/3;
		desiredDist = 800;
		double maxExcapeAnge = maxEscapeAngleForSpeed(wToSurf.getSpeed());
		double maxBodyRot = 10./180.0*Math.PI;
		double probeStick = 220;
		double angleToCorrectDist = 20.0/180.0*Math.PI;
		double testAngle = 0;
		//double headOnAngle = wToSurf.source.pos.angleTo( wToSurf.target.pos );
		Point curV = new Point( bot.velocity );
		botStatus futBot = new botStatus();
		futBot.pos = new Point(bot.pos);
		futBot.velocity = new Point(bot.velocity);
		futBot.heading = bot.heading;
		futBot.speed = bot.speed;
		double pDanger = 0;
		if ( futBot.speed < 0 ) {
			// normalizing heading and speed
			futBot.speed *= -1;
			futBot.heading += Math.PI;
		}
		for (int cnt=1; cnt <= timeToSurf; cnt++) {
			double latVel = wToSurf.source.pos.lateralVelocityOf( futBot );
			double radVel = wToSurf.source.pos.radialVelocityOf( futBot );
			double distToSource = wToSurf.source.pos.dist(futBot.pos);
			double speed = futBot.velocity.length();
			maxBodyRot = (10-0.75*Math.abs(speed))/180.0*Math.PI;
			double direction = futBot.velocity.angle();
			//dbg("-----------------");
			//dbg("curBotPos " + futBot.pos + " curBotVel = " + futBot.velocity);
			//dbg("direction = " + direction + " velocity based");
			if ( speed == 0 ) {
				direction = futBot.heading;
				//dbg("direction = " + direction + " heading based");
			}
			//dbg(" old velocity " + futBot.velocity  + " speed = " + futBot.velocity.length());
			//dbg("dist to source = " + distToSource + " radVel " +  radVel + " latVel " + latVel);
			double rA = maxBodyRot*Math.signum( latVel);
			//dbg(" angle to futBot.pos " + wToSurf.source.pos.angleTo( futBot.pos ) );
			double headOnAngle = wToSurf.source.pos.angleTo( futBot.pos );
			double desiredHeading =  surfDirection*Math.PI/2; // relative to head on
			//dbg("desiredHeading raw " + desiredHeading);
			double ramDangerDist = 100;
			double ramMultiplier =1;
			boolean isEnemyRamming =  isRamming(futBot, wToSurf);
			desiredHeading += 
				Math.signum( distToSource - desiredDist ) 
				* angleToCorrectDist * surfDirection 
				* Math.min(1, Math.abs(distToSource - desiredDist)/36);
			if ( isEnemyRamming ) {
				desiredHeading = surfDirection * Math.PI*40./180. ;
				//dbg("enemy ramming");
			}
			//dbg("desiredHeading enemy dist corrected " + desiredHeading/Math.PI*180);
			double probeAngleIncrement = surfDirection*1.0/180*Math.PI;
			desiredHeading -= probeAngleIncrement;
			Point testP = null;
			int cntHere=0, cntMax=200;
			do {
				desiredHeading += probeAngleIncrement; // relative to head on
				testP = new Point ( futBot.pos);
				testP.addPolarVec( 4+futBot.speed/8*probeStick, desiredHeading + headOnAngle );
				//dbg("desiredHeading probe in the box " + desiredHeading*180/Math.PI + " to point " + testP);
			} while ( !testP.isInsideBox( SWcorner, NEcorner) && (cntMax > cntHere++)  );
			//dbg("desiredHeading box corrected " + desiredHeading*180/Math.PI);
			desiredHeading += headOnAngle;
			double missingAngle = Utils.normalRelativeAngle( desiredHeading - futBot.heading );

			//dbg(" velocity before zero correction " + futBot.velocity  + " speed = " + futBot.velocity.length() + " heading " + futBot.heading);
			if ( Utils.isNear(speed,0) && (Math.abs( missingAngle ) > Math.PI/2) ) {
				// we have a chance to choose right direction
				futBot.heading += Math.PI;
			}
			missingAngle = Utils.normalRelativeAngle( desiredHeading - futBot.heading );
			//dbg(" velocity before rotation " + futBot.velocity  + " speed = " + futBot.velocity.length() + " heading " + futBot.heading);

			//dbg("need to rotate " + missingAngle );
			rA = Math.min(maxBodyRot, Math.abs(missingAngle)) *Math.signum( missingAngle );
			if ( Math.abs( missingAngle ) > Math.PI/2 ) {
				// we need to decelerate and rotate other way
				rA = -rA;
			}
			futBot.velocity.rotateBy( rA );
			futBot.heading+= rA;
			//dbg(" new velocity after rotation " + futBot.velocity  + " speed = " + futBot.velocity.length() + " heading " + futBot.heading);

			Point updVelocity = futBot.velocity;
			missingAngle = Utils.normalRelativeAngle( desiredHeading - futBot.heading );
			if ( Math.abs( missingAngle) <= Math.PI/2 ) {
				// may be we can accelerate
				if ( speed < 8 ) {
					double maxAccel = 1;
					speed += maxAccel;
					speed = Math.min( speed, 8 );
					updVelocity = new Point();
					updVelocity.addPolarVec( speed, futBot.heading);
					//dbg("latVel after accel " + wToSurf.source.pos.lateralVelocityOf( futBot) );
				}
			} else {
				// we should flip velocity direction 
				double maxDeAccel = 2;
				speed -= maxDeAccel;
				if (speed < 0) {
					speed *= -1;
					futBot.heading += Math.PI;
				}
				updVelocity = new Point();
				updVelocity.addPolarVec( speed, futBot.heading);
			}
			futBot.velocity = updVelocity;
			futBot.speed = futBot.velocity.length();

			//dbg(" new velocity after accel/deaccel " + futBot.velocity  + " speed = " + futBot.velocity.length());
			testP = new Point ( futBot.pos);
			testP.add( futBot.velocity );

			double dxToEnemy = Math.abs( enemy.pos.x - futBot.pos.x );
			double dyToEnemy = Math.abs( enemy.pos.y - futBot.pos.y );
			boolean collisionWithEnemy = false;
			if ( (dxToEnemy < (34.0))  && (dyToEnemy < (34.0) ) ){
				collisionWithEnemy = true;
			}
			if (testP.isInsideBox( SWcorner, NEcorner) && !collisionWithEnemy) {
				if ( isEnemyRamming && (Math.PI-Math.abs(desiredHeading - headOnAngle))< 36/distToSource ) {
					//dbg("dive in protection: not going that direction");
				} else {
					futBot.pos = testP;
				}
			} else {
				futBot.velocity = new Point(0,0);
				futBot.speed = 0;
			}
			//dbg("futPos " + futBot.pos + " curV " + futBot.velocity + " speed " + futBot.velocity.length() );
			// let's get the danger of this point
			headOnAngle = wToSurf.source.pos.angleTo( wToSurf.target.pos );
			double firingAngle = Utils.normalRelativeAngle( wToSurf.source.pos.angleTo( futBot.pos ) - headOnAngle );
			int gfBin = xToBinInRange( firingAngle, -maxExcapeAnge, maxExcapeAnge, wToSurf.Nbins );
			pDanger = wToSurf.waveDanger[gfBin];
			//if (distToSource > 60 ) pDanger =0;
			//pDanger += 1e-2/wToSurf.source.pos.dist( futBot.pos ); //bot danger
			//dbg("fut Pos " + futBot.pos + " curV " + futBot.velocity + " speed " + futBot.velocity.length() + " with danger " + pDanger + " gfBin " + gfBin + " at gF " + firingAngle/maxExcapeAnge);
			if (isEnemyRamming && !wToSurf.isReal ) {
				pDanger =0; // when enemy is ramming we should be more worried about getting away
			}
			if ( distToSource < 100 ) {
				pDanger =0; // if source is so close there is no wave to dodge the bullet
			}
			// adding some enemy distance dependent danger
			pDanger *= 1.0/distToSource;
			pDanger += .1/distToSource;
			//dbg("this point danger = " + pDanger);
			if ( pDanger <= safestDanger ) {
				safestDanger = pDanger;
				destP = futBot.pos;
				//dbg("safest point <=-- " +  destP + " with gf = " + firingAngle/maxExcapeAnge);
			}
		}
		DangerPoint dest = new DangerPoint();
		dest.pos = destP;
		dest.danger = pDanger;
		return dest;
	}

	public class DangerPoint {
		Point pos = new Point(0,0);
		double danger = 0;
		public String toString() {
			String str="";
			str += pos + " with danger " + danger;
			return str;
		}
	}

	public boolean isRamming( botStatus bot, Wave wToSurf ) {
		return isRamming( bot, wToSurf.source );
	}

	public boolean isRamming( botStatus bot, botStatus enemy ) {
		if (bot.pos.dist(enemy.pos) < ramAvoidDistance ) {
			return true;
		}
		return false;
	}

	public void moveToDestination( Point destP ) {
		double angleToDest = bot.pos.angleTo( destP );
		double distToDest = bot.pos.dist( destP );
		double realHeading = bot.heading;
		double rotAngle = Utils.normalRelativeAngle( angleToDest - realHeading );
		double direction = 1;
		if ( Math.abs(rotAngle) > Math.PI/2 ) {
			rotAngle = Utils.normalRelativeAngle( rotAngle - Math.PI );
			direction = -1;

		}
		setTurnLeftRadians( rotAngle);
		setAhead(distToDest*direction);
	}

	public void initTic() {
		time = getTime();
		//dbg("tic: " + time );
		updateMasterBot();
		//dbg(bot);
		updateMyWaves();
		updateEnemyWaves();
		if (hitsByEnemyTree.size() > MaxVirtualWavesEnemy) enemyEmitsVirtualWaves=false;
		if ( enemy != null && enemyEmitsVirtualWaves ) {
			// virtual wave from enemy
			boolean isReal = false;
			addEnemyWave( enemy, lastEnemyBulletEnergy, isReal );
		}
	}

	public void updateMasterBot() {
		double oldAverageSpeed = 0;
		if ( bot != null ) {
			if ( time == bot.time ) {
				// already update for this tic
				// most likely we get it during the onScannedRobot event
				return;
			}
			oldAverageSpeed = bot.averageSpeed;
		}
		botStatus prevBotStatus = bot;
		bot = new botStatus();
		bot.prev = prevBotStatus;
		bot.name = getName();
		bot.pos.x=getX();
		bot.pos.y=getY();
		bot.gunHeat = getGunHeat();
		// proper Cartesian heading
		bot.heading = Utils.normalRelativeAngle(Math.PI/2 -  getHeadingRadians());
		bot.speed = getVelocity();
		double weight=0.9;
		bot.averageSpeed = weight*oldAverageSpeed + (1-weight)*bot.speed; 
		bot.time = getTime();
		bot.velocity.x = bot.speed * Math.cos( bot.heading );
		bot.velocity.y = bot.speed * Math.sin( bot.heading );
	}

	public void updateMyWaves() {
		for (int i = 0; i < myWaves.size(); i++) {
			Wave w = (Wave) myWaves.get(i);
			if ( w.isCrossing( enemy.pos ) ) {
				//dbg("wave fired at time " + w.source.time + " crosses bot " + enemy.name);
				boolean isTrueHit = false;
				registerHitOnMyWave(w, enemy.pos, isTrueHit);
			}
			if ( w.isBehind(enemy) ) {
				for (String gunName : w.firingSolutions.keySet() ) {
					if ( !gunStatsMap.containsKey( gunName ) ) {
						dbg("ERROR: unknown gun = " + gunName + " is used in the wave");
						continue;
					} else {
						gunStatsMap.get(gunName).firedCnt++;
						if (w.isReal)
							gunStatsMap.get(gunName).realFiredCnt++;
					}
				}
				myWaves.remove(i); 
				i--;
			}
		}
	}

	public void registerHitOnMyWave( Wave w, Point pos, boolean isTrueHit) {
		Hit _hit = new Hit();
		_hit.hitAngle =  w.hitAngle( pos );
		_hit.waveCount = w.count;
		_hit.isWaveReal= w.isReal;
		_hit.isReal = isTrueHit;
		double[] coord = w.source.treeCoord( w.target, w.bulletEnergy );
		hitsByMeTree.addPoint( coord, _hit);
		if ( w.isReal ) {
			realWaveHitsByMeTree.addPoint( coord, _hit);
		}
		/*
		// we have lateral symmetry
		// MAKE SURE ALL TREES ARE UPDATED
		double[] refl_coord = w.source.treeCoordReflected( coord );
		hitsByMeTree.addPoint( refl_coord, _hit);
		*/

		boolean isHitFiringSolutionFound = false;
		double enemyAngularSize = 25.0/w.source.pos.dist( pos );
		ArrayList<String> gunsToRemove = new ArrayList<String>();
		for (String gunName : w.firingSolutions.keySet() ) {
			if ( !gunStatsMap.containsKey( gunName ) ) {
				dbg("ERROR: unknown gun = " + gunName + " is used in the wave");
				continue;
			} else {
				double firingAngle = w.firingSolutions.get( gunName );
				if ( Math.abs( firingAngle - _hit.hitAngle ) < enemyAngularSize ) {
					// this firing solution has a hit
					isHitFiringSolutionFound = true;
					gunStatsMap.get(gunName).firedCnt++;
					gunStatsMap.get(gunName).hitCnt++;
					if (isTrueHit || w.isReal) {
						// FIXME: double counting is possible
						gunStatsMap.get(gunName).realHitCnt++;
						gunStatsMap.get(gunName).realFiredCnt++;
					}
					gunsToRemove.add( gunName ); // otherwise we double counting
				}
			}
		}
		if ( isTrueHit && !isHitFiringSolutionFound ) {
			dbg("SHAME: did not find firing solution to match true hit at time " + time);
			notMatchedFiringSolutionCnt[getRoundNum()]++;
		}
		for (String gunName : gunsToRemove ) {
			w.firingSolutions.remove( gunName ); 
		}
	}

	public void registerHitOnEnemyWave( Wave w, Point pos, boolean isTrueHit) {
		Hit _hit = new Hit();
		_hit.hitAngle =  w.hitAngle( pos );
		_hit.waveCount = w.count;
		_hit.isWaveReal= w.isReal;
		_hit.isReal = isTrueHit;
		double[] coord = w.source.treeCoord( w.target, w.bulletEnergy );
		hitsByEnemyTree.addPoint( coord, _hit);
		if ( w.isReal ) {
			realWaveHitsByEnemyTree.addPoint( coord, _hit);
		}

		/*
		// we have lateral symmetry
		double[] refl_coord = w.source.treeCoordReflected( coord );
		_hit = new Hit();
		_hit.hitAngle =  -w.hitAngle( pos );
		_hit.waveCount = w.count;
		_hit.isWaveReal= w.isReal;
		_hit.isReal = isTrueHit; // this is derived hit so think carefully
		hitsByEnemyTree.addPoint( refl_coord, _hit);
		if ( w.isReal ) {
			realWaveHitsByEnemyTree.addPoint( refl_coord, _hit);
		}
		*/
	}

	public void updateEnemyWaves() {
		for (int i = 0; i < enemyWaves.size(); i++) {
			Wave w = (Wave) enemyWaves.get(i);
			if (
					w.isCrossing( bot.pos ) 
					&& virtrualHitOnRealWave // shall we track for seen by enemy angles
			   ) { 
				//dbg("wave fired at time " + w.source.time + " crosses bot " + bot.name);
				boolean isTrueHit = false;
				registerHitOnEnemyWave( w, bot.pos, isTrueHit );
			}
			if ( w.isBehind( bot ) ) {
				enemyWaves.remove(i); 
				i--;
			}
		}
	}

	public void onScannedRobot(ScannedRobotEvent e) {
		time = getTime();
		enemyLastSeenTime = time;
		//dbg("enemy Scanned at time " + time);
		updateMasterBot();
		//dbg("my Pos" + bot);

		double enemyHeading = Utils.normalAbsoluteAngle( getHeadingRadians() + e.getBearingRadians());
		double distance = e.getDistance();
		double oldAverageSpeed = 0;
		double oldEnergy = 0;
		double enemyBulletEnergy = 0;
		if ( enemy != null ) {
			oldAverageSpeed = enemy.averageSpeed;
			oldEnergy = enemy.energy;
		}
		botStatus prevEnemyStatus = enemy;
		enemy = new botStatus();
		enemy.prev = prevEnemyStatus;
		enemy.energy = e.getEnergy();
		enemy.pos.x = bot.pos.x + Math.sin(enemyHeading) * distance;
		enemy.pos.y = bot.pos.y + Math.cos(enemyHeading) * distance;
		enemy.name=e.getName();
		// gun heat calculations
		if ( enemy.prev != null ) {
			enemy.gunHeat = Math.max(0, enemy.prev.gunHeat - getGunCoolingRate());
		} else {
			if (time < 30 ) {
				enemy.gunHeat = 3.0 - getGunCoolingRate()*time;
			}
		}
		// proper Cartesian heading
		enemy.heading =  gameAngleToCartesian(e.getHeadingRadians());
		enemy.speed = e.getVelocity();
		double weight = 0.9;
		enemy.averageSpeed = weight*oldAverageSpeed + (1-weight)*enemy.speed; 
		enemy.velocity.x = enemy.speed * Math.cos( enemy.heading );
		enemy.velocity.y = enemy.speed * Math.sin( enemy.heading );
		enemy.time = time;
		//dbg(enemy);


		// let's decide if enemy fired a bullet or it is something else
		double energyChange = oldEnergy - enemy.energy;
		if ( energyChange > 0 ) {
			if ( prevEnemyStatus != null ) {
				// could it be collision with a wall?
				double accel = Math.abs( enemy.speed ) - Math.abs( prevEnemyStatus.speed );
				//dbg( "accel = " + accel);
				if (accel < - Rules.DECELERATION) {
					// collision with a wall or a bot
					if (!enemy.pos.isInsideBox( SWcornerCollision, NEcornerCollision)) {
						// collision with a wall
						double wallCollisionPenalty = Rules.getWallHitDamage( prevEnemyStatus.speed );
						energyChange -= wallCollisionPenalty;
						//dbg("tic: " + time + " wall hit detected with energy penalty " + wallCollisionPenalty + " for the speed " + prevEnemyStatus.speed );
					} else {
						if ( Utils.isNear(energyChange, Rules.ROBOT_HIT_DAMAGE) ) {
							// bots collided
							energyChange -= Rules.ROBOT_HIT_DAMAGE; // 0.6 energy points
						} else {
							dbg("this is likely a bot with bot collision on top of the wall hit with energy change of " + energyChange);
							dbg("TODO make code to handle such collisions");
						}
					}
				}
			}
			if (  energyChange > Rules.MAX_BULLET_POWER ) {
				dbg(" unaccounted energy drop forcing energy of bullet to " + Rules.MAX_BULLET_POWER );
				energyChange = Rules.MAX_BULLET_POWER;
			}
			if ( !Utils.isNear(energyChange, 0) ) {
				enemyBulletEnergy = energyChange;
				if ( energyChange < Rules.MIN_BULLET_POWER ) {
					dbg(" suspiciously low energy drop of " + energyChange + " for a bullet");
				}
				//dbg("tic: " + time + " enemy fired bullet with energy " + enemyBulletEnergy);
				boolean isReal = true;
				enemy.gunHeat = enemyBulletEnergy/5.0 + 1.0 - getGunCoolingRate(); // we detect 1 click after the fire
				addEnemyWave( enemy, enemyBulletEnergy, isReal );
			}
		}

		// simple radar lock
		double radRotAngle =  Utils.normalRelativeAngleDegrees(getHeading() + e.getBearing() - getRadarHeading());
		if (radRotAngle > 0) {
			radRotAngle += 45/2;
		} else {
			radRotAngle -= 45/2;
		}
		setTurnRadarRight( radRotAngle);

		// Fire the gun. Remember firing happens with previous turn solutions
		// so we do it before calculating the new one
		if ( fSols.containsKey( bestGunName ) ) {
			firingAngle = fSols.get( bestGunName );
		} else {
			firingAngle = 0;  // always point to enemy
			bulletEnergy = Double.NaN;
		}
		_bullet = null;
		bulletEnergy = Math.min( getEnergy()-.05, bulletEnergy);
		if (bulletEnergy >= Rules.MIN_BULLET_POWER && !Double.isNaN(bulletEnergy) ) {
			if ( Utils.isNear(0, Math.abs(getGunTurnRemainingRadians())) || ( gunSkippingFireWhenReadyCnt >2 && !isBulletShielder ) ) {
				//dbg("Gun angle = " + gameAngleToCartesian(getGunHeadingRadians()) );
				//dbg("Gun remaining angle = " + Math.abs(getGunTurnRemainingRadians()) );
				gunSkippingFireWhenReadyCnt=0;
				_bullet = setFireBullet( bulletEnergy );
			} else {
				if ( Utils.isNear(bot.gunHeat,0) ) {
					gunSkippingFireWhenReadyCnt++;
					//dbg("not firing when cold at tic " + time + " heat is " + getGunHeat());
				   	notFiredWhenCold[getRoundNum()]++;
				}
			}
			if (_bullet != null ) {
				//dbg("==> bullet fired at time " + time + " with energy " + _bullet.getPower());
				firedBulletCnt++;
				bulletEnergy = _bullet.getPower();
				String gunName = "realGun";
				if ( !gunStatsMap.containsKey( gunName ) ) {
					GunStats _gs = new GunStats(); gunStatsMap.put(gunName, _gs);
				}
				double realGunFiringAngle =  Utils.normalRelativeAngle( gameAngleToCartesian(getGunHeadingRadians()) - bot.pos.angleTo( enemy.pos ) ); // with respect to HoT
				fSols.put(gunName, realGunFiringAngle);
			}
			if ( hitsByMeTree.size() <= MaxVirtualWavesMy ) {
				Wave w = new Wave();
				myWaveCnt++;
				w.count = myWaveCnt;
				w.source = bot;
				w.target = enemy;
				if ( !Double.isNaN(bulletEnergy) ) {
					bulletEnergy = IDEAL_BULLET_ENERGY; // default energy. FIXME: use distance
				}
				w.bulletEnergy = bulletEnergy;
				w.firingSolutions = fSols;
				w.isReal = false;
				if (_bullet != null ) {
					w.isReal = true;
					w.bulletEnergy = _bullet.getPower();
					gunStatsMap.get( bestGunName ).realUseCnt++;
				}
				myWaves.add(w);
				for (Wave eW: enemyWaves ) {
					eW.addSafetyCorridorFromWave( w );
				}
				if ( bestGunName.equals( "shieldGun") ) {
					// for shield gun we also fire virtual bullet
					w = new Wave();
					myWaveCnt++;
					w.count = myWaveCnt;
					w.source = bot;
					w.target = enemy;
					w.bulletEnergy = IDEAL_BULLET_ENERGY; // default energy. FIXME: use distance
					w.firingSolutions = fSols;
					w.isReal = false;
					myWaves.add(w);
				}
			} else {
				dbg("hitsByMeTree is full, do not generate wave which cannot process");
			}
		}

		double distToEnemy = bot.pos.dist( enemy.pos );
		// gun firing angle
		if ( distToEnemy < 200 ) {
			bulletEnergy = robocode.Rules.MAX_BULLET_POWER;
		} else {
			bulletEnergy = IDEAL_BULLET_ENERGY;
		}
		//bulletEnergy = (IDEAL_BULLET_ENERGY/1.95)*3*(300)/(distToEnemy);
		bulletEnergy = Math.min( bulletEnergy, robocode.Rules.MAX_BULLET_POWER );
		bulletEnergy = (Math.round( bulletEnergy * 10 ) - .5 )/10; // exploiting x.x5 power bug in old bots

		bulletEnergy = Math.min( bulletEnergy, enemy.energy/4); // damage per bullet
		bulletEnergy = Math.max( 0.102, bulletEnergy); // energy below 0.1 not allowed
		if ( isBulletShielder ) {
			bulletEnergy = 0.3; // bullet shielder
		}
		bulletEnergy = Math.min( getEnergy()-.05, bulletEnergy);

		// create enough info about future bot status
		botStatus fbot = new botStatus();
		fbot.time = bot.time+1;
		fbot.pos = new Point(bot.pos );
		fbot.pos.add( bot.velocity );
		fbot.velocity = new Point( bot.velocity );

		// some useful defaults
		double requiredGunPrecision = gunPrecision;
		Point targetFuturePos = new Point( enemy.pos ); 
		targetFuturePos.add( enemy.velocity );

		if ( !isBulletShielder ) {
		fSols = calcFiringSolutions( fbot, enemy, bulletEnergy );
		double bestHitRate = Double.NEGATIVE_INFINITY;
		firingAngle = 0;
		for(String gunName : fSols.keySet() ) {
			if ( !gunStatsMap.containsKey( gunName ) ) {
				//dbg("Adding new gun " + gunName +" in gunStatsMap " + gunStatsMap);
				GunStats _gs = new GunStats(); gunStatsMap.put(gunName, _gs);
			}
			GunStats _gs = gunStatsMap.get( gunName );
			double _hitRate = _gs.getHitRatio();
			if (_gs.realFiredCnt > 50 ) {
				_hitRate = _gs.getRealHitRatio();
			}
			if ( _hitRate > bestHitRate ) {
				if ( gunName == "RandomGun" && firedBulletCnt < 100 ) {
					// do nothing we do not have enough stats for the RandomGun
				} else {
					bestHitRate = _hitRate;
					firingAngle = fSols.get( gunName );
					bestGunName = gunName;
				}
			}
		}
		//dbg("Best gun " + bestGunName);
		
		// anti bullet shield measure: avoid headOn firing
		double enemyBotAngularSize = 18/distToEnemy;
		if ( Math.abs(firingAngle) < (enemyBotAngularSize)/20 ) {
			//dbg("avoiding exact headOn firing");
			if (Math.random() > 0.5) {
				firingAngle =  enemyBotAngularSize/5;
			} else {
				firingAngle = -enemyBotAngularSize/5;
			}
		}
			requiredGunPrecision = Math.max(gunPrecision, enemyBotAngularSize/3);
		} else {
			// we are using bullet shielder

			// let's find the closest wave
			Wave wToIntercept = null;
			double waveClosestDist = Double.NEGATIVE_INFINITY;
			for (int i = 0; i < enemyWaves.size(); i++) {
				Wave w = (Wave) enemyWaves.get(i);
				if (!w.isReal) continue; // not intercepting virtual waves
				double surfDist = w.getSurfingDistanceTo(fbot.pos);
				// positives mean that wave is behind the point
				if (surfDist > -(18+2*w.getSpeed()) ) {
					// this wave either behind or too close
					// so too late to worry about it
					continue;
				} else {
					if ( surfDist > waveClosestDist ) {
						// this wave is closer to us
						waveClosestDist = surfDist;
						wToIntercept = w;
					}
				}
			}
			if ( wToIntercept == null ) {
				bulletEnergy = 0; // fake bullet which gun will not fire
				firingAngle = 0;
			} else {
				dbg("At time " + time + " trying to intercept wave fired at " + wToIntercept.source.time);
				// creating  rudimentary intercepting wave
				Wave w = new Wave();
				w.bulletEnergy = bulletEnergy;
				w.source = fbot;
				w.target = null;

				double enemyBulletGuessedAngle = 0; // relative to HoT
				if (wToIntercept.surfingInfo.isCalculated) {
					double maxEscapeAngle = maxEscapeAngleForSpeed( wToIntercept.getSpeed());
					int i=wToIntercept.surfingInfo.safestReachableBin;
					enemyBulletGuessedAngle = binToXInRange( i, -maxEscapeAngle, maxEscapeAngle, wToIntercept.waveDanger.length);
				}
				Point firingSolution = counterBulletFiringAngle( wToIntercept, enemyBulletGuessedAngle, w);
				firingAngle = firingSolution.x;
				//requiredGunPrecision = firingSolution.y;
				bulletEnergy = firingSolution.y;
				firingAngle = Utils.normalRelativeAngle( firingAngle - fbot.pos.angleTo( enemy.pos ) ); // relative to HoT
				if ( Double.isNaN(firingAngle) ) {
					firingAngle = 0;
					bulletEnergy = Double.NaN; // we should not fire such bullet
				}

				String gunName = "shieldGun";
				bestGunName = gunName;
				if ( !gunStatsMap.containsKey( gunName ) ) {
					//dbg("Adding new gun " + gunName +" in gunStatsMap " + gunStatsMap);
					GunStats _gs = new GunStats(); gunStatsMap.put(gunName, _gs);
				}
				fSols.put(gunName, firingAngle);
			}
		}

		// gun Rotation
		double futureGunFiringAngle =  Utils.normalRelativeAngle( gameAngleToCartesian(getGunHeadingRadians()) - fbot.pos.angleTo( targetFuturePos ) ); // with respect to HoT
		double remainGunTurn = firingAngle - futureGunFiringAngle;
		setTurnGunLeftRadians(1.0 * remainGunTurn);
	}

	public HashMap<String, Double> calcFiringSolutions( botStatus bot, botStatus enemy, double bulletEnergy) {
		double[] coord = bot.treeCoord( enemy, bulletEnergy );
		boolean isSequentialSorting = true;
		double maxExcapeAnge = maxEscapeAngleForSpeed(Rules.getBulletSpeed( bulletEnergy ));
		double halfBotAngularSize = 25/bot.pos.dist( enemy.pos );
		int Nbins = (int) Math.ceil( 2*maxExcapeAnge/ halfBotAngularSize * 3.5);

		HashMap<String, Double> fSols = new HashMap<String, Double>();

		// simple non learning guns
		double firingAngle = 0;
		fSols.put("HoT", firingAngle);

		firingAngle = 2*(Math.random()-0.5)*maxExcapeAnge;
		fSols.put("RandomGun", firingAngle);

		// kd-tree guns
		KdTree.WeightedManhattan<Hit> tree;
		tree =  hitsByMeTree;
		//tree =  realWaveHitsByMeTree;
		List<KdTree.Entry<Hit>> cluster = tree.nearestNeighbor( coord, neigborsNumForMasterBotGun, isSequentialSorting );

		TargetingWeights tw = null;

		// Main Gun
		tw = new TargetingWeights();
		tw.realWave_realHit = 1; tw.realWave_virtualHit = 1;
		tw.virtualWave_realHit = 1; tw.virtualWave_virtualHit = 1;
		tw.timeDecay = Double.POSITIVE_INFINITY;

		tw.smoothingFactor = 0.05;
		firingAngle = calcFiringAngleFromClusterForTargetWeight( cluster, tw, maxExcapeAnge, Nbins);
		fSols.put("MainGunSM0.05", firingAngle);

		tw.smoothingFactor = 0.10;
		firingAngle = calcFiringAngleFromClusterForTargetWeight( cluster, tw, maxExcapeAnge, Nbins);
		fSols.put("MainGunSM0.10", firingAngle);

		// Anti Surfer Gun 
		tw = new TargetingWeights();
		tw.realWave_realHit = 0.5; tw.realWave_virtualHit = 1;
		tw.virtualWave_realHit = 0.1; tw.virtualWave_virtualHit = 0.1;
		tw.timeDecay = 10000;

		tw.smoothingFactor = 0.05;
		firingAngle = calcFiringAngleFromClusterForTargetWeight( cluster, tw, maxExcapeAnge, Nbins);
		fSols.put("AntiSurferGun0.05", firingAngle);

		tw.smoothingFactor = 0.10;
		firingAngle = calcFiringAngleFromClusterForTargetWeight( cluster, tw, maxExcapeAnge, Nbins);
		fSols.put("AntiSurferGun0.10", firingAngle);

		tw.smoothingFactor = 0.05;
		tw.timeDecay = 5000;
		firingAngle = calcFiringAngleFromClusterForTargetWeight( cluster, tw, maxExcapeAnge, Nbins);
		fSols.put("AntiSurferGun0.05d5000", firingAngle);


		tw.smoothingFactor = 0.10;
		tw.timeDecay = 5000;
		firingAngle = calcFiringAngleFromClusterForTargetWeight( cluster, tw, maxExcapeAnge, Nbins);
		fSols.put("AntiSurferGun0.10d5000", firingAngle);

		return fSols;
	}

	public double calcFiringAngleFromClusterForTargetWeight(
			List<KdTree.Entry<Hit>> cluster, TargetingWeights tw,
			double maxExcapeAnge, int Nbins
			) {
		if  (cluster.size() != 0 ) {
			//dbg("===================================");
			//dbg("cluster size = " + cluster.size() + " tree size = " + tree.size());
			//dbg("coord: " + Arrays.toString(coord) );
			double normAngle=0;
			double[] hitProb = new double[Nbins];


			for ( KdTree.Entry<Hit> hitEntry : cluster ) {
				double hitWeight = 1e-2/(1e-2 + hitEntry.distance);
				//double timeContribution = ( myWaveCnt - hitEntry.value.waveCount)/myStatsDecayTime;
				double timeContribution = ( myWaveCnt - hitEntry.value.waveCount)/tw.timeDecay;
				timeContribution = Math.exp( - timeContribution*timeContribution );
				hitWeight *= timeContribution;

				if (hitEntry.value.isWaveReal) { 
					if ( hitEntry.value.isReal ) {
						hitWeight *= tw.realWave_realHit;
					} else {
						hitWeight *= tw.realWave_virtualHit;
					}
				} else {
					if ( hitEntry.value.isReal ) {
						hitWeight *= tw.virtualWave_realHit;
						dbg("Error: real hit cannot be on virtual wave");
					} else {
						hitWeight *= tw.virtualWave_virtualHit;
					}
				}
				int bin = xToBinInRange(hitEntry.value.hitAngle, -maxExcapeAnge, maxExcapeAnge, Nbins);
				hitProb[bin] += hitWeight;

				normAngle += hitWeight*hitEntry.value.hitAngle; // averaged angle
				//dbg("Entry v = " + hitEntry.value.hitAngle + " w = " + hitWeight + " dist = " + hitEntry.distance);
			}
			//dbg("normalized angle " + normAngle);
			//firingAngle = normAngle;
			hitProb = smoothWaveDanger( hitProb, tw.smoothingFactor); 
			//dbg( arrayToTextPlot(hitProb) + " <--- firing angle weights");
			int bestBin = (int) (Nbins -1)/2;;
			for(int i=1; i< Nbins; i++) {
				if ( hitProb[i] > hitProb[bestBin] ) {
					bestBin = i;
				}
			}
			double[] hitChoice = new double[Nbins];
			//int bestBin = xToBinInRange(firingAngle, -maxExcapeAnge, maxExcapeAnge, Nbins);
			hitChoice[bestBin] = 1;
			firingAngle = binToXInRange( bestBin, -maxExcapeAnge, maxExcapeAnge, Nbins);
			//dbg( arrayToTextPlot(hitChoice) + " <--- firing angle choice");
		} else {
			//dbg("cluster is empty");
			firingAngle = 0;
		}
		return firingAngle;
	}

	public double[] smoothWaveDanger(double[] danger, double spreadFraction) {
		// spread fraction tells the fraction of the bins to which danger spills
		int Nbins = danger.length;
		int dw = (int) Math.max(2, spreadFraction*Nbins);
		double[] smoothedDanger = new double[Nbins];
		//dbg( arrayToTextPlot(danger) + " before smoothing");
		for (int i=0; i< Nbins; i++) {
			double hitWeight = danger[i];
			for (int k= 0; k<=4*dw; k++) {
				double hitSpreadWeight = 1.0/(1.0+1.0*k*k/dw/dw);
				if ((i-k) >= 0 )   smoothedDanger[i-k] += hitWeight*hitSpreadWeight;
				if (k==0) continue;
				if ((i+k) < Nbins) smoothedDanger[i+k] += hitWeight*hitSpreadWeight;
			}
		}
		//dbg( arrayToTextPlot(smoothedDanger) + " after smoothing");
		return smoothedDanger;
	}

		public Point counterBulletFiringAngle(Wave enemyWave, double enemyBulletHeading, Wave counterFireWave) {
			// we will use point to get firing angle and precision
			Point firingSolution = new Point(Double.NaN,0); 
			enemyBulletHeading += enemyWave.source.pos.angleTo( enemyWave.target.pos ); // adding HoT angle
			dbg("enemyBulletHeading = " + enemyBulletHeading);
			Point enemyBulletPosition = new Point ( enemyWave.source.pos );
			Point bulletVelocity = new Point(0,0);
		   	bulletVelocity.addPolarVec(enemyWave.getSpeed(), enemyBulletHeading); 
			long gapTime = counterFireWave.source.time - enemyWave.source.time;
			if ( gapTime < 0 ) {
				dbg("ERROR: do not know yet how to counter fire before enemy fired");
				return firingSolution;
			}
			// let's calculate where is the bullet at the counter fire time
			while ( gapTime > 0 ) {
				gapTime--;
				enemyBulletPosition.add ( bulletVelocity );
			}
			if ( enemyWave.getDistTo(enemyBulletPosition) > enemyWave.getDistTo(counterFireWave.source.pos) ) {
				dbg("WARNING: this wave is behind us already, no need to fire at it");
				return firingSolution;
			}
			int maxCnt = 100;
			for (long i = 1; i <= maxCnt; i++) {
				enemyBulletPosition.add ( bulletVelocity );
				if ( enemyWave.getDistTo(enemyBulletPosition) > enemyWave.getDistTo(counterFireWave.source.pos) ) {
					dbg("WARNING: this wave is behind us already, no need to fire at it");
					return firingSolution;
				}
				double overShotDist = i*counterFireWave.getSpeed() - counterFireWave.getDistTo( enemyBulletPosition );
				if ( overShotDist > 0 ) {
					// our wave just crossed the enemy bullet
					double overShotTime = overShotDist/counterFireWave.getSpeed();
					Point idealInterseptPoint = new Point( enemyBulletPosition );
					idealInterseptPoint.x -= bulletVelocity.x * 0.5; // half a step back
					idealInterseptPoint.y -= bulletVelocity.y * 0.5;
					double firingAngle = counterFireWave.source.pos.angleTo( idealInterseptPoint );
					double idealInterceptingBulletSpeed = counterFireWave.source.pos.dist( idealInterseptPoint )/ (i-0.5);
					dbg("ideal speed is = " + idealInterceptingBulletSpeed + " current is " + counterFireWave.getSpeed() + " travel time " + i + " intercept time is " + (counterFireWave.source.time + i) );
					double idealInterceptingBulletEnergy = (20 - idealInterceptingBulletSpeed)/3.0;
					dbg("required energy for bullet " + idealInterceptingBulletEnergy);
					double realisticInterceptingBulletEnergy = putInRange( idealInterceptingBulletEnergy, Rules.MIN_BULLET_POWER, Rules.MAX_BULLET_POWER);
					double realisticInterceptingBulletSpeed = Rules.getBulletSpeed(realisticInterceptingBulletEnergy);
					dbg("adjusted energy " + realisticInterceptingBulletEnergy);
					double bulletCentersMismatch = Math.abs(idealInterceptingBulletSpeed - Rules.getBulletSpeed(realisticInterceptingBulletEnergy) )*(i-0.5);
					dbg("bulletCentersMismatch " + bulletCentersMismatch + " for bullet speed " + Rules.getBulletSpeed( realisticInterceptingBulletEnergy ));
					if ( bulletCentersMismatch > realisticInterceptingBulletSpeed/4) {
						// we have not find good solution
						dbg("WARNING: the intercepting solution is likely to miss, discarding it");
						return firingSolution; 
					}

					firingSolution.x = firingAngle;
					firingSolution.y = realisticInterceptingBulletEnergy;
					return firingSolution;
				}
			}

			dbg("WARNING: could not intercept the enemy wave with heading angle " + enemyBulletHeading );
			return firingSolution;
		}

	public void onBulletHit(BulletHitEvent event) {
		// master bot hit someone
		hitBulletCnt++;
		//dbg("my hit ratio " + hitBulletCnt + "/" + firedBulletCnt + " = " + String.format( "%.1f", 100.0*hitBulletCnt/firedBulletCnt) + " %");
		//printStats();
		// let's account for this energy change in enemy
		Bullet bullet = event.getBullet();
		double enemyEnergyChange = 4*bullet.getPower();
		enemyEnergyChange += (bullet.getPower() > 1) ? 2*( bullet.getPower() -1 ) : 0;
		//dbg( "tic: " + time + " enemy lost " + enemyEnergyChange + " due to master bot fire");
	    enemy.energy -= enemyEnergyChange;

		boolean isTrueHit = true;
		boolean hitWaveFound = false;
		Point pos = new Point(bullet.getX(),bullet.getY());
		for (int i = 0; i < myWaves.size(); i++) {
			Wave w = (Wave) myWaves.get(i);
			if ( !w.isReal ) { continue; }
			if ( w.isCrossing( pos ) ) {
				//dbg("wave fired at time " + w.source.time + " crosses bot " + enemy.name);
				hitWaveFound = true;
				registerHitOnMyWave(w, enemy.pos, isTrueHit);
			}
		}
		if (!hitWaveFound) {
			dbg("SHAME: did not find wave containing bullet at tic " + time);
		}
	}

	public void onHitByBullet(HitByBulletEvent event) {
		// master bot was hit by someone
		enemyHitBulletCnt++;
		//dbg("enemy hit ratio " + enemyHitBulletCnt + "/" + enemyFiredBulletCnt + " = " + String.format( "%.1f", 100.0*enemyHitBulletCnt/enemyFiredBulletCnt) + " %");
		//printStats();
		// let's find which wave hits me
		Wave wHit = null;
		Bullet bullet = event.getBullet();
		// account for enemy energy gain
		double enemyEnergyChange = 3*bullet.getPower();
	    enemy.energy += enemyEnergyChange;
		//dbg( "tic: " + time + " enemy gain " + enemyEnergyChange + " due to its successful fire, enemy energy now " + enemy.energy);
		Point hitLocation = new Point( bullet.getX(), bullet.getY() );
		wHit = findWaveClosestToPoint( hitLocation, bullet.getPower() );
		if ( wHit != null) {
			// update hit stats
			wHit.isReal = true;
			boolean isTrueHit = true;
			registerHitOnEnemyWave( wHit, hitLocation, isTrueHit);
			updateEnemyWavesDanger();
		}
	}

	public void onBulletHitBullet(BulletHitBulletEvent event) {
		// ("bullets intersection detected");
		bulletHitBulletCnt[getRoundNum()]++;
		Wave wHit = null;
		Bullet bullet = event.getHitBullet(); // enemy bullet
		// TODO this is approximation of the intersection location 
		// but this is ok for hitAngle detection
		Point hitLocation = new Point( bullet.getX(), bullet.getY() );
		wHit = findWaveClosestToPoint( hitLocation, bullet.getPower() );
		if ( wHit != null) {
			// update hit stats
			wHit.isReal = true;
			boolean isTrueHit = true;
			registerHitOnEnemyWave( wHit, hitLocation, isTrueHit);
			// now let's make this wave virtual
			// so we do not act against a wave with no bullet
			wHit.isReal = false;
			updateEnemyWavesDanger();
		}
	}

	public void updateEnemyWavesDanger() {
		for (int i = 0; i < enemyWaves.size(); i++) {
			Wave w = (Wave) enemyWaves.get(i);
			assignWaveDanger(w);
		}
	}

	public Wave findWaveClosestToPoint( Point hitLocation, double bulletEnergy ) {
		Wave wHit = null;
		double closestWaveDist = Double.POSITIVE_INFINITY;
		for (int i = 0; i < enemyWaves.size(); i++) {
			Wave w = (Wave) enemyWaves.get(i);
			//dbg("dist from wave to hitLocation is " +w.getSurfingDistanceTo( hitLocation ) );
			//dbg( "wave bullet " + w.bulletEnergy + " struck bullet " + bullet.getPower() );
			if ( ! Utils.isNear(w.bulletEnergy, bulletEnergy) ) {
				continue;
			}
			double d = Math.abs( w.getSurfingDistanceTo( hitLocation ) );
			if ( d < closestWaveDist ) {
				closestWaveDist = d;
				wHit = w;
			}
		}
		if ( closestWaveDist > 25 ) { // bot diagonal
			dbg("Code imperfection detected: ==================");
			dbg("Shame on me. Could not find enemy wave which struck me at " + hitLocation);
			if ( wHit != null ) {
				dbg( "wave bullet " + wHit.bulletEnergy + " struck bullet " + bulletEnergy );
				dbg("The closest surf of the wave is at distance " + closestWaveDist );
			}
			dbg("==============================================");
			wHit = null;
		}
		return wHit;
	}

	public void onRoundEnded(RoundEndedEvent event) {
		// RoundEndedEvent has priority higher than WinEvent or DeathEvent
		// so we cannot use it to print unifying stats :(, stats would be one round late
		//dbg("Round ended");
		//printStats();
	}

	public void onWin(WinEvent event) {
		winCnt++;
		dbg("Master bot win the round");
		printStats();
	}
	public void onDeath(DeathEvent event) {
		dbg("Master bot loose the round");
		printStats();
	}

	public void printStats() {
		double[] hitRatios = {1.0*hitBulletCnt/firedBulletCnt, 0,  1.0*enemyHitBulletCnt/enemyFiredBulletCnt};
		String out = "";
		out += "hit ratios my vs enemy: " + arrayToTextPlot(hitRatios) ;
		out += "  ";
		out += "my hit ratio " + hitBulletCnt + "/" + firedBulletCnt + " = " + String.format( "%.1f", 100.0*hitBulletCnt/firedBulletCnt) + " %";
		out += "  ";
		out += "enemy hit ratio " + enemyHitBulletCnt + "/" + enemyFiredBulletCnt + " = " + String.format( "%.1f", 100.0*enemyHitBulletCnt/enemyFiredBulletCnt) + " %";
		TreeSet<String>  guns = new TreeSet<String>();
		guns.addAll( gunStatsMap.keySet() );
		for( String gunName : guns ) {
			GunStats _gs = gunStatsMap.get(gunName);
			out += "\n";
			out += String.format( "%20s", gunName) + " " + _gs.toString();
		}
		//
		out += "\n";
		out += "Counts of not found matching firing solution: " + Arrays.toString(notMatchedFiringSolutionCnt);
		out += "\n";
		out += "Counts of not firing when gun is cold       : " + Arrays.toString(notFiredWhenCold);
		out += "\n";
		out += "Counts of bullet hit bullet                 : " + Arrays.toString(bulletHitBulletCnt);
		out += "\n";
		out += "Win ratio " + winCnt + "/" + (1+getRoundNum());
	
		dbg( out );
		//dbg("my hits tree size " + hitsByMeTree.size() );
		//dbg("enemy hits tree size " + hitsByEnemyTree.size() );
	}

	public void addEnemyWave( botStatus enemy, double bulletEnergy, boolean isReal ) {
		// enemy is status at time of the fire detection 
		// according to robocode physics this is 2 tics later
		// after actual fire
		// see robocode physics
		// http://robowiki.net/wiki/Robocode/Game_Physics
		//dbg("detected enemy " + enemy.name + " fire at time " + time + " with bullet energy " + bulletEnergy);
		Wave w = new Wave();
		enemyWaveCnt++;
		w.count = enemyWaveCnt;
		w.isReal = isReal;
		if ( w.isReal ) enemyFiredBulletCnt++;
		w.source = enemy;
		if ( enemy.prev != null ) {
			w.source = enemy.prev;
		}
		w.target = bot;
		if ( bot.prev != null ) {
			w.target = bot.prev;
			if ( bot.prev.prev != null ) {
				w.target = bot.prev.prev;
			}
		}
		w.bulletEnergy = bulletEnergy;
		lastEnemyBulletEnergy=w.bulletEnergy;
		enemyWaves.add(w);

		assignWaveDanger(w);
		if (w.isReal) {
			// add safety corridors from the flying waves of master bot 
			for (Wave mW: myWaves ) {
				w.addSafetyCorridorFromWave(mW);
			}
		}

		//if ( w.isReal ) {
		//dbg( Arrays.toString(w.waveDanger));
		//dbg( arrayToTextPlot(w.waveDanger));
		//}
	}

	public void assignWaveDanger( Wave w ) {
		// wave danger
		double[] coord = w.source.treeCoord( w.target, w.bulletEnergy );
		double maxExcapeAnge = maxEscapeAngleForSpeed(w.getSpeed());
		double distToTarget = w.source.pos.dist( w.target.pos );
		double halfBotAngularSize = 25.0/distToTarget;
		int Nbins = (int) Math.ceil( 2*maxExcapeAnge/ halfBotAngularSize * 3.5);
		w.waveDanger = new double[Nbins];
		w.Nbins=Nbins;

		boolean isSequentialSorting = true;
		double enemyHitRate = 1.0 * enemyHitBulletCnt / (enemyFiredBulletCnt+1);
		//KdTree.WeightedManhattan<Hit> tree = hitsByEnemyTree;
		KdTree.WeightedManhattan<Hit> tree = realWaveHitsByEnemyTree;
		int nn = (int) (100 + Math.sqrt( tree.size()));

		TargetingWeights tw = new TargetingWeights();
		if ( enemyHitRate < .10 ) {
			// This is anti fixed gun measure, i.e. HoT, linear, circular, etc
			tw.realWave_realHit = 10; tw.realWave_virtualHit = -0.02;
			tw.virtualWave_realHit = 0; tw.virtualWave_virtualHit = 0;
			tw.timeDecay = Double.POSITIVE_INFINITY;
			tw.smoothingFactor = halfBotAngularSize/maxExcapeAnge/5;
		} else {
			// enemy likely to use GF stats
			tw.realWave_realHit = 10; tw.realWave_virtualHit = 1;
			tw.virtualWave_realHit = 1; tw.virtualWave_virtualHit = 1;
			tw.timeDecay = Double.POSITIVE_INFINITY;
			tw.smoothingFactor = halfBotAngularSize/maxExcapeAnge/5;
		}

		List<KdTree.Entry<Hit>> cluster = tree.nearestNeighbor( coord, nn, isSequentialSorting );
		if  (cluster.size() != 0 ) {
			//dbg("===================================");
			//dbg("cluster size = " + cluster.size() + " tree size = " + tree.size());
			//dbg("coord: " + Arrays.toString(coord) );
			for ( KdTree.Entry<Hit> hitEntry : cluster ) {
				double rawHitWeight = 1e-2/(1e-2 + hitEntry.distance);
				double hitWeight = rawHitWeight;
				//dbg("Entry v = " + hitEntry.value.hitAngle + " w = " + hitWeight + " dist = " + hitEntry.distance);
				int bin = xToBinInRange(hitEntry.value.hitAngle, -maxExcapeAnge, maxExcapeAnge, w.Nbins);

				double timeContribution = ( myWaveCnt - hitEntry.value.waveCount)/tw.timeDecay;
				timeContribution = Math.exp( - timeContribution*timeContribution );
				hitWeight *= timeContribution;

				if (hitEntry.value.isWaveReal) { 
					if ( hitEntry.value.isReal ) {
						hitWeight *= tw.realWave_realHit;
					} else {
						hitWeight *= tw.realWave_virtualHit;
					}
				} else {
					if ( hitEntry.value.isReal ) {
						hitWeight *= tw.virtualWave_realHit;
					} else {
						hitWeight *= tw.virtualWave_virtualHit;
					}
				}
				w.waveDanger[bin]+=hitWeight;
			}
			// smoothing Wave
			//dbg( arrayToTextPlot(w.waveDanger) + " before smoothing");
			double[] smoothedDanger = smoothWaveDanger( w.waveDanger, tw.smoothingFactor);
			//dbg( arrayToTextPlot(smoothedDanger) + " after smoothing");
			w.waveDanger = smoothedDanger;
		}
	}

	public class TargetingWeights {
		public double 
			 realWave_realHit = 1, realWave_virtualHit = 1,
			 virtualWave_realHit = 1, virtualWave_virtualHit = 1,
		// note: virtualWave_realHit is actually impossible
			 timeDecay = Double.POSITIVE_INFINITY,
			 smoothingFactor = 0.1;
	}

	public int xToBinInRange( double x, double lowEnd, double hiEnd, int Nbins ) {
		if (lowEnd > hiEnd ) {
			double tmp = hiEnd;
			hiEnd = lowEnd;
			lowEnd = tmp;
		}
		if ( x <=lowEnd ) return 0;
		if ( x >= hiEnd  ) return (Nbins - 1);
		double range = hiEnd - lowEnd;
		return (int) Math.round((x-lowEnd)/range*(Nbins - 1));
	}

	public double binToXInRange( int bin, double lowEnd, double hiEnd, int Nbins ) {
		if (lowEnd > hiEnd ) {
			double tmp = hiEnd;
			hiEnd = lowEnd;
			lowEnd = tmp;
		}
		if ( bin <= 0 ) return lowEnd;
		if ( bin >= (Nbins -1) ) return hiEnd;
		double range = hiEnd - lowEnd;
		return lowEnd + range*bin/(Nbins -1);
	}

	public static String arrayToTextPlot( double[] bins) {
		// outputs text style/plot histograms
		// inspired by https://github.com/holman/spark
		String ticks = "▁▂▃▄▅▆▇█";
		int ticksNum = 8;
		double bMax = Double.NEGATIVE_INFINITY;
		double bMin = Double.POSITIVE_INFINITY;
		int ind;
		double b;
		for (int i=0; i < bins.length; i++) {
			// find stats
			b=bins[i];
			if ( b > bMax)
				bMax = b;
			if ( b < bMin)
				bMin = b;
		}
		String sout = "";
		double range = bMax - bMin;
		if ( range==0 )
			range = 1;
		for (int i=0; i < bins.length; i++) {
			// normalize
			b=(bins[i]-bMin)/range;
			ind = (int) Math.floor( b*ticksNum );
			if ( ind == ticksNum )
				ind = ticksNum - 1;
			sout += ticks.charAt(ind);
		}
		sout += " With range from " + bMin + " to " + bMax;
		return sout;
	}
	public void dbg( Object str ) {
		System.out.println( str );
	}

	public class Point extends Point2D.Double {
		//double x=0;
		//double y=0;

		public Point() {x=0; y=0;}
		public Point( double xs, double ys) {x=xs; y=ys;}
		public Point( Point pOther) {x=pOther.x; y=pOther.y;}

		public String toString() { return ("P=( " + x +", " + y + " )"); };
		public double dist(Point p) {
			double dx = x - p.x;
			double dy = y - p.y;
			return Math.sqrt(dx*dx+dy*dy);
		}
		public Point vectorToTarget(Point p) {
			Point vec = new Point();
			vec.x = p.x-x;
			vec.y = p.y-y;
			return vec;
		}
		public double angleTo(Point p) {
			double dx = p.x - x;
			double dy = p.y - y;
			return Math.atan2( dy, dx);
		}
		public double angle() {
			// gives its own direction angle
			Point origin = new Point();
			return origin.angleTo( this );
		}
		public double scalarProduct(Point p) {
			return ( x*p.x + y*p.y );
		}
		public double vectorProduct(Point p) {
			// self x p
			return ( x*p.y - y*p.x );
		}
		public double length() {
			return dist( new Point() );
		}
		public Point rotateBy(double a) {
			// rotate vector presented in this point by angle a
			double s=Math.sin(a);
			double c=Math.cos(a);
			double xo = x;
			double yo = y;
			x=c*xo - s*yo;
			y=s*xo + c*yo;
			return this;
		}
		public Point add(Point p) {
			// add vector presented by point p
			x+=p.x;
			y+=p.y;
			return this;
		}
		public Point addPolarVec(double r, double angle) {
			// add vector presented by point p
			x+=r*Math.cos( angle );
			y+=r*Math.sin( angle );
			return this;
		}
		public double lateralVelocityOf(botStatus target) {
			// positive means that target is moving CCW
			Point radVec = this.vectorToTarget( target.pos );
			double latVel = radVec.vectorProduct( target.velocity);
			latVel /= radVec.length();
			return latVel;
		}

		public double radialVelocityOf(botStatus target) {
			// positive means that target is moving away
			Point radVec = this.vectorToTarget( target.pos );
			double radVel = radVec.scalarProduct( target.velocity);
			radVel /= radVec.length();
			return radVel;
		}
		public boolean isInsideBox(Point c1, Point c2) {
			// check if point is inside rectangle defined by corners c1 and c2
			return (           ( Math.min( c1.x, c2.x) < x)
					&& ( Math.max( c1.x, c2.x) > x)
					&& ( Math.max( c1.y, c2.y) > y)
					&& ( Math.min( c1.y, c2.y) < y) );
		}

	}

	public class botStatus {
		Point pos = new Point();
		botStatus prev = null;
		double energy = 0;
		double heading = 0;
		double speed = 0;
		double averageSpeed = 0;
		double gunHeat = 0;
		String name = "";
		long time = 0;
		Point velocity = new Point();

		double [] treeCoord(botStatus target, double bulletEnergy) {
			double[] coord = new double[myGunKdTreeDims];
			double[] scale = {1./1000, 1./8, 1./8, 1./3, 1./8} ;
			double dist = pos.dist( target.pos );
			coord[0] = dist;
			coord[1] = pos.lateralVelocityOf(target);
			coord[2] = pos.radialVelocityOf(target);
			coord[3] = bulletEnergy;
			coord[4] = Math.abs(target.averageSpeed);
			
			for(int i=0; i < coord.length; i++) {
				coord[i] *= scale[i];
			}

			//dbg("coord " +arrayToTextPlot(coord));
			return coord;
		}

		double [] treeCoordReflected(double[] coord) {
			// take in account symmetry with respect to lateral velocity
			double[] reflected_coord = new double[coord.length];
			for (int i=0; i < coord.length; i++) {
				reflected_coord[i] =coord[i];
			}
			reflected_coord[1] = -coord[1]; // lateral velocity flip
			return reflected_coord;
		}

		public String toString() {
			String str="";
			str += name;
			str += " status at time ";
			str += time;
			str += ":";
			str += " position ";
			str += pos;
			str += ", speed ";
			str += speed;
			str += ", averageSpeed ";
			str += averageSpeed;
			str += ", heading ";
			str += heading;
			str += ", velocity ";
			str += velocity;
			return str;
		}
	}

	public class SurfingInfo {
		public boolean isCalculated = false;
		int safestReachableBin=0;
		Point destinationPoint = null;
	}

	public class Wave {
		botStatus source = new botStatus();
		botStatus target = new botStatus();
		long count = 0;
		boolean isReal = false; // virtual or real wave
		int Nbins = 31; // keep it odd, so the head-on-angle is the proper bin
		double[] waveDanger = new double[Nbins];
		double bulletEnergy = 0;
		HashMap<String, Double> firingSolutions = null;
		public SurfingInfo surfingInfo = new SurfingInfo();
		public ArrayList<Point> safetyCorridors = new ArrayList<Point>();

		public Wave() {
			int gf0bin = (int)((Nbins-1)/2);
			double gf0danger = 0*1e-4;
			double width=Nbins*.2;
			waveDanger[ gf0bin ] = gf0danger; // small danger to head-On-Angle
			for (int i=0; i< Nbins; i++ ) {
				double diff = i-gf0bin;
				diff /= width;
				waveDanger[i] = gf0danger/( 1 + diff*diff);
			}

		}

		double getSpeed() {
			return ( 20 - bulletEnergy * 3 );
		}

		double getDistTraveled() {
			long t = time - source.time;
			double dist = t*getSpeed();
			return dist;
		}
		double getDistTo(Point p) {
			double dist = source.pos.dist( p);
			return dist;
		}
		double getSurfingDistanceTo(Point p) {
			// positive means that wave already passed the point p
			return getDistTraveled() -  getDistTo( p );
		}
		boolean isOutside() {
			if ( getDistTraveled() >  BattleFieldDiagonal ) {
				return true;
			}
			return false;
		}
		boolean isBehind(botStatus b) {
			if ( getSurfingDistanceTo(b.pos) >  25 ) {
				return true;
			}
			return false;
		}
		boolean isCrossing( Point p ) {
			// check the wave crosses a bot at point p 
			if ( Math.abs( getSurfingDistanceTo(p) ) < 18 ) {
				return true;
			}
			return false;
		}
		double hitAngle(Point p) {
			double hit = source.pos.angleTo( p );
			double headOn = source.pos.angleTo( target.pos );
			return Utils.normalRelativeAngle( hit - headOn );
		}
		public double firingAngleWithMaxDanger() {
			// relative to HoT direction
			double maxEscapeAngle = maxEscapeAngleForSpeed(getSpeed());
			return binToXInRange( maxBinInArray( waveDanger ), -maxEscapeAngle, maxEscapeAngle, waveDanger.length);
		}
		public Point calcSafetyCorridorFromWave(Wave w) {
			//Point.x is CCW end and Point.y is CW end of safety corridor (sC)
			Point sC = null;
			Point noSolution = null;
			double headOnAngle = w.source.pos.angleTo( w.target.pos );
			if ( w.firingSolutions == null ) return noSolution;
			if ( !w.firingSolutions.containsKey( "realGun" ) ) return noSolution;
			double bulletHeading = headOnAngle + w.firingSolutions.get( "realGun" );

			Point bulletVelocity = new Point(0,0);
		   	bulletVelocity.addPolarVec(w.getSpeed(), bulletHeading); 


			long gapTime = this.source.time - w.source.time;
			long this_delta = 0, w_delta = 0;
			if (gapTime > 0 ) {
				w_delta = gapTime;
			} else {
				this_delta = -gapTime;
			}

			// search for wave interception time
			Point bulletHeadPos;
			double distToBulletHead;
			double thisWaveTravel;
			long maxCnt=100; // max tics to propagate 
			int i=-1;
			do {
				i++;
				bulletHeadPos = new Point(w.source.pos);
				bulletHeadPos.addPolarVec( w.getSpeed()*(w_delta+i), bulletHeading );
				distToBulletHead = this.getDistTo( bulletHeadPos );
				thisWaveTravel = this.getSpeed()*(this_delta+i);
			} while ( distToBulletHead > thisWaveTravel && i < maxCnt );

			// i=0; wave is already behind the bullet
			// i>=maxCnt waves did not crossed within allocated clicks
			if ( i == 0 || i >= maxCnt ) return noSolution;

			// if we are here this wave crosses the other wave bullet within given time

			// very primitive safety corridor assignment
			headOnAngle = this.source.pos.angleTo( this.target.pos );
			double sC_angle1 = this.source.pos.angleTo( bulletHeadPos ) - headOnAngle;
			Point bulletTailPos = new Point(w.source.pos);
			bulletTailPos.addPolarVec( w.getSpeed()*(w_delta+i-1), bulletHeading );
			double sC_angle2 = this.source.pos.angleTo( bulletTailPos ) - headOnAngle;

			sC_angle1 = Utils.normalRelativeAngle( sC_angle1 );
			sC_angle2 = Utils.normalRelativeAngle( sC_angle2 );
			if ( sC_angle1 > sC_angle2 ) {
				sC = new Point(sC_angle1, sC_angle2); 
			} else {
				sC = new Point(sC_angle2, sC_angle1); 
			}
			return sC;
		}

		public double getBotAngularSizeAtPoint(Point p) {
			double dist = getDistTo( p );
			return 2*Math.atan( 25.0/dist);
		}

		public void addSafetyCorridorFromWave( Wave w ) {
			if (!this.isReal) return;
			if (!w.isReal) return;
			Point sC = calcSafetyCorridorFromWave(w);
			if (sC == null) return;
			safetyCorridors.add(sC);

			// if we are adding safety corridor, let's decrease wave danger in it
			double maxEscapeAngle = maxEscapeAngleForSpeed( getSpeed() );
			if ( 
					!isInRange( sC.x, maxEscapeAngle, maxEscapeAngle ) &&
					!isInRange( sC.y, maxEscapeAngle, maxEscapeAngle )
			   )
		   	{
				// at least one end of the safety corridor is within maxEscapeAngle 
				int gfBinMin = xToBinInRange( sC.y, -maxEscapeAngle, maxEscapeAngle, this.waveDanger.length );
				int gfBinMax = xToBinInRange( sC.x, -maxEscapeAngle, maxEscapeAngle, this.waveDanger.length );
				// first, assign 0 danger at the shadow points
				for(int i=gfBinMin; i<= gfBinMax; i++) {
					if (waveDanger[i] > 0) {
						waveDanger[i] *= 0.0;
					}
				}
				// second, smooth it by the bot width
				double[] oldWaveDanger  = waveDanger.clone();
				double botHalfWidth = (int) Math.ceil(getBotAngularSizeAtPoint(target.pos)/(2*maxEscapeAngle)*waveDanger.length/2.0); // in units of bins number
				for(int i=gfBinMin; i<= gfBinMax; i++) {
					for (int k = (int)Math.max(0, i-botHalfWidth); k<(int)Math.min(i+botHalfWidth, waveDanger.length); k++) {
						waveDanger[i] += oldWaveDanger[k]/(2*botHalfWidth);
					}
				}

				
				surfingInfo.isCalculated = false; // we need to update surfing solutions
			}
		}

		public void onPaint(Graphics2D g) {
			// default colors (enemy waves)
			Color colorSafestReachableGF = new Color(0x00, 0xff, 0x00, 0x80);
			Color colorWaveReal    = new Color(0xdd, 0x00, 0x00, 0x80);
			Color colorWaveVirtual = new Color(0x00, 0x00, 0xdd, 0x80);
			if ( source.name.equals( getName() ) ) {
				// colors for my bot waves
				colorWaveReal    = new Color(0xdd, 0xdd, 0x00, 0x80);
				colorWaveVirtual = new Color(0x00, 0x00, 0xdd, 0x80);
			}
			Color colorWave = colorWaveReal;
			if ( !isReal ) {
				colorWave = colorWaveVirtual;
				return;
			} else {
				colorWave = colorWaveReal;
			}
			double headOnAngle = source.pos.angleTo( target.pos );
			double distTraveled = (time - source.time)*getSpeed();

			// draw present position
			g.setColor( colorWave );
			drawCircle(g, source.pos, distTraveled);
			// draw next tic position
			g.setColor( colorWave.brighter() );
			drawCircle(g, source.pos, distTraveled+getSpeed());

			// draw source pos at fire time
			g.setColor( colorWave );
			//drawRect(g, source.pos, 36, 36);

			// draw target pos at fire time
			g.setColor( colorWave );
			//drawRect(g, target.pos, 36, 36);

			// draw HoT line
			g.setColor( colorWave );
			//drawLine(g, source.pos, target.pos);

			// draw firing solutions
			if ( firingSolutions != null ) {
				for (String gunName : firingSolutions.keySet() ) {
					if ( !gunName.equals("realGun") ) continue;
					double fA = headOnAngle + firingSolutions.get(gunName);
					Point endP = new Point( source.pos );
					endP.addPolarVec( distTraveled + getSpeed(), fA);
					g.setColor( colorWave );
					//drawLine(g, source.pos, endP);
				}
			}

			// draw wave danger
			g.setColor( colorWave );
			double maxEscapeAngle = maxEscapeAngleForSpeed(getSpeed());
			double maxDanger = waveDanger[ maxBinInArray( waveDanger ) ];
			if (maxDanger == 0 ) maxDanger=1;
			for( int i=0; i< waveDanger.length; i++) {
				double fA = binToXInRange( i, -maxEscapeAngle, maxEscapeAngle, waveDanger.length);
				fA += headOnAngle;

				Point strtP = new Point( source.pos );
				strtP.addPolarVec( distTraveled , fA);
				Point endP = new Point( source.pos );
				endP.addPolarVec( distTraveled + waveDanger[i]/maxDanger*getSpeed(), fA);

				g.setColor( colorWave );
				drawLine(g, strtP, endP);
			}

			if (surfingInfo.isCalculated) {
				// safest reachable GF
				int i = surfingInfo.safestReachableBin;
				double fA = binToXInRange( i, -maxEscapeAngle, maxEscapeAngle, waveDanger.length);
				fA += headOnAngle;
				Point strtP = new Point( source.pos );
				//strtP.addPolarVec( distTraveled , fA);
				Point endP = new Point( source.pos );
				endP.addPolarVec( distTraveled, fA);
				g.setColor( colorSafestReachableGF );
				drawLine(g, strtP, endP);

				// show expected safe position for a bot at the surf time
				g.setColor( colorSafestReachableGF );
				drawRect(g, surfingInfo.destinationPoint, 36, 36);
			}

			// drawing safety corridors
			for( Point sC : safetyCorridors ) {
				double fA1 = headOnAngle + sC.x;
				double fA2 = headOnAngle + sC.y;

				g.setColor( colorSafestReachableGF );
				Point strtP = new Point( source.pos );
				strtP.addPolarVec( distTraveled, fA1);
				Point endP = new Point( source.pos );
				endP.addPolarVec( distTraveled, fA2);
				
				drawLine(g, strtP, endP);
			}
		}
	}

	public class Hit {
		public double hitAngle=0;
		public long waveCount=0;
		public boolean isReal = false;
		public boolean isWaveReal = false;
	}

	public class GunStats {
		public int firedCnt=0, hitCnt=0;
		public int realFiredCnt=0, realHitCnt=0;
		public int realUseCnt=0;
		public double getHitRatio() {
			if (firedCnt > 0) {
				return 1.0*this.hitCnt/this.firedCnt;
			}
			return 0;
		}
		public double getRealHitRatio() {
			if (firedCnt > 0) {
				return 1.0*this.realHitCnt/this.realFiredCnt;
			}
			return 0;
		}
		public String toString(){
			String out="";
			out +=  "all waves hit ratio " + this.hitCnt + "/" + this.firedCnt + " = " + String.format( "%.1f", 100.0*getHitRatio()) + "%";
			out += " ";
			out +=  "real waves hit ratio " + this.realHitCnt + "/" + this.realFiredCnt + " = " + String.format( "%.1f", 100.0*getRealHitRatio()) + "%";
			out += " real use count " + realUseCnt;
			return out;
		}
	}

	public double gameAngleToCartesian( double angle ) {
		return Utils.normalRelativeAngle(Math.PI/2 - angle);
	}

	public void onPaint(Graphics2D g) {
		time = getTime();
		g.setColor(new Color(0x00, 0xff, 0x00, 0x80));
		// draw my waves
		for (int i = 0; i < myWaves.size(); i++) {
			Wave w = (Wave) myWaves.get(i);
			w.onPaint(g);
		}
		// draw enemy waves
		for (int i = 0; i < enemyWaves.size(); i++) {
			Wave w = (Wave) enemyWaves.get(i);
			w.onPaint(g);
		}
	}

	public int maxBinInArray( double[] a ) {
		int maxBin = (int)((a.length-1)/2); // gf=0 bin is the default max
		for (int i=0; i< a.length; i++ ) {
			if ( a[maxBin] < a[i]) {
				maxBin = i;
			}
		}
		return maxBin;
	}
	public double putInRange(double x, double min, double max) {
		if (x< min) return min;
		if (x> max) return max;
		return x;
	}

	public boolean isInRange(double x, double min, double max) {
		if (x< min) return false;
		if (x> max) return false;
		return true;
	}

	public double maxEscapeAngleForSpeed(double bulletSpeed) {
		return Math.asin(Rules.MAX_VELOCITY/bulletSpeed);
	}

	// Graphic utilities
	public static void drawLine( Graphics2D g, Point2D.Double strtP, Point2D.Double endP ) {
		g.drawLine((int)strtP.x, (int)strtP.y, (int)endP.x, (int)endP.y);
	}
	public static void drawCircle( Graphics2D g, Point2D.Double p, double R ) {
		g.drawOval( (int)(p.x - R), (int)(p.y - R), (int)(2*R), (int)(2*R) );
	}
	public static void drawRect( Graphics2D g, Point2D.Double p, double width, double height ) {
		g.drawRect( (int)(p.x - width/2), (int)(p.y - height/2), (int)(width), (int)(height) );
	}

};

// vim: tabstop=4: shiftwidth=4
